/*
 * Copyright (c) 2001-2005 The Regents of The University of Michigan
 * Copyright (c) 2010 Advanced Micro Devices, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met: redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer;
 * redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution;
 * neither the name of the copyright holders nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Steve Reinhardt
 *          Nathan Binkert
 */

#include <cassert>

#include "base/callback.hh"
#include "base/inifile.hh"
#include "base/match.hh"
#include "base/misc.hh"
#include "base/trace.hh"
#include "base/types.hh"
#include "debug/Checkpoint.hh"
#include "sim/sim_object.hh"
#include "sim/stats.hh"
#include <string>

using namespace std;


////////////////////////////////////////////////////////////////////////
//
// SimObject member definitions
//
////////////////////////////////////////////////////////////////////////

//
// static list of all SimObjects, used for initialization etc.
//
SimObject::SimObjectList SimObject::simObjectList;
SimObject::SimObjectList SimObject::simCPUList;
SimObject::SimObjectList SimObject::simSwitchCPUList;


//
// SimObject constructor: used to maintain static simObjectList
//
SimObject::SimObject(const Params *p)
    : EventManager(p->eventq), _params(p)
{
#ifdef DEBUG
    doDebugBreak = false;
#endif

    simObjectList.push_back(this);
    //cprintf("SIMOBJ LIST push back %s\n", p->name);

    state = Running;

    if((p->name.find("basem5.cpu")!=string::npos)&&(p->name.find('.',7)==string::npos))
    {
    	//cprintf("BASE cpu out for %s\n", p->name);
    	simCPUList.push_back(this);
    }
    else if((p->name.find("o3m5.cpu")!=string::npos)&&(p->name.find('.',7)==string::npos))
    {
    	//cprintf("SWITCH cpu for %s\n", p->name);
    	simSwitchCPUList.push_back(this); 
    }
}

void
SimObject::init()
{
}

//
// static function: call init() on all SimObjects.
//

void
initAll()
{
	SimObject::init_All();
}

void
SimObject::init_All()
{
    SimObjectList::iterator i = simObjectList.begin();
    SimObjectList::iterator end = simObjectList.end();

    for (; i != end; ++i) {
        //cprintf("init_All for %s\n", (*i)->name());
        SimObject *obj = *i;
        obj->init();
    }
}


void
SimObject::loadState(Checkpoint *cp)
{
    if (cp->sectionExists(name())) {
        DPRINTF(Checkpoint, "unserializing\n");
        unserialize(cp, name());
    } else {
        DPRINTF(Checkpoint, "no checkpoint section found\n");
    }
}

void
SimObject::initState()
{
}

void
SimObject::startup()
{
}

void
SimStartup()
{
	SimObject::Sim_Startup();
	
}

void
SimObject::Sim_Startup()
{
    SimObjectList::iterator i = simObjectList.begin();
    SimObjectList::iterator end = simObjectList.end();

    for (; i != end; ++i) {
        //cprintf("initStats for %s\n", (*i)->name());
	(*i)->initState();
    }

    for (i = simObjectList.begin(); i != end; ++i) {
        //cprintf("startup calls for %s\n", (*i)->name());
        (*i)->startup();
    }

}


//
// no default statistics, so nothing to do in base implementation
//
void
SimObject::regStats()
{
}

void
SimObject::printStats(usage_counts &totalusage)
{
}

void
SimObject::regFormulas()
{
}

//
// static function: call resetStats() on all SimObjects.
//

void
SimObject::resetAllStats()
{
    SimObjectList::iterator i = simObjectList.begin();
    SimObjectList::iterator end = simObjectList.end();

    for (; i != end; ++i) {
        (*i)->resetStats();
    }
}

struct SimObjectResetCB : public Callback
{
    virtual void process() { SimObject::resetAllStats(); }
};

namespace {
    static SimObjectResetCB StatResetCB;
}

void
regAllStats()
{
	SimObject::reg_AllStats();
}


void
SimObject::reg_AllStats()
{
/*
    SimObjectList::iterator i = simObjectList.begin();
    SimObjectList::iterator end = simObjectList.end();


    string substr("cache");
    size_t found;

    for (; i != end; ++i) {
	found=((*i)->name()).find(substr);
	if(found==string::npos)
	{
        	//cprintf("registering stats for %s\n", (*i)->name());
        	(*i)->regStats();
        	//(*i)->regFormulas();
	}
    }

    Stats::registerResetCallback(&StatResetCB);
*/
}

void
SimObject::SST_FF()
{
    SimObjectList::iterator i = simObjectList.begin();
    SimObjectList::iterator end = simObjectList.end();


    size_t found;

    for (; i != end; ++i) {
    	//cprintf("FF SST SST for %s\n", (*i)->name());
        found=((*i)->name()).find(".system");
        if(found!=string::npos)
        {
                cprintf("\nChanging system to timing mode for %s\n", (*i)->name());
                (*i)->setMemoryMode(Enums::timing);  //timing mode
        }
    }

    SimObjectList::iterator j=simSwitchCPUList.begin();

    for (i=simCPUList.begin(); i!=simCPUList.end(); ++i, ++j) 
    {
        cprintf("Switching cpu from %s to %s\n", (*i)->name(), (*j)->name());
        (*i)->switchOut();
        (*j)->takeOverFrom((BaseCPU*)(*i));
    }

    for (i=simObjectList.begin(); i != end; ++i) {
        //cprintf("FF SST SST for %s\n", (*i)->name());
        found=((*i)->name()).find(".system");
        if(found!=string::npos)
        {
                cprintf("Resume system %s\n", (*i)->name());
                (*i)->resume(); 
        }
    }

}


void
SimObject::printROIStats(usage_counts &totalusage)
{
    SimObjectList::iterator i = simObjectList.begin();
    SimObjectList::iterator end = simObjectList.end();

    for (; i != end; ++i) {
    	if((*i)->name().find("basem5.cpu")==string::npos)
        	//cprintf("pirnt stats for %s\n", (*i)->name());
        	(*i)->printStats(totalusage);
    }
}


void
SimObject::printAllStats(usage_counts &totalusage)
{
    SimObjectList::iterator i = simObjectList.begin();
    SimObjectList::iterator end = simObjectList.end();

    /**
     * @todo change cprintfs to DPRINTFs
     */

    for (; i != end; ++i) {
                //cprintf("pirnt stats for %s\n", (*i)->name());
                (*i)->printStats(totalusage);
    }
}


void
SimObject::resetStats()
{
}

//
// static function: serialize all SimObjects.
//
void
SimObject::serializeAll(ostream &os)
{
    SimObjectList::reverse_iterator ri = simObjectList.rbegin();
    SimObjectList::reverse_iterator rend = simObjectList.rend();

    for (; ri != rend; ++ri) {
        SimObject *obj = *ri;
        obj->nameOut(os);
        obj->serialize(os);
   }
}


#ifdef DEBUG
//
// static function: flag which objects should have the debugger break
//
void
SimObject::debugObjectBreak(const string &objs)
{
    SimObjectList::const_iterator i = simObjectList.begin();
    SimObjectList::const_iterator end = simObjectList.end();

    ObjectMatch match(objs);
    for (; i != end; ++i) {
        SimObject *obj = *i;
        obj->doDebugBreak = match.match(obj->name());
   }
}

void
debugObjectBreak(const char *objs)
{
    SimObject::debugObjectBreak(string(objs));
}
#endif

unsigned int
SimObject::drain(Event *drain_event)
{
    state = Drained;
    return 0;
}

void
SimObject::resume()
{
    state = Running;
}

/*
void
SimObject::setMemoryMode(State new_mode)
{
    panic("setMemoryMode() should only be called on systems");
}
*/


void
SimObject::setMemoryMode(Enums::MemoryMode mem_mode)
{
}


void
SimObject::switchOut()
{
//    panic("Unimplemented!");
}

void
SimObject::takeOverFrom(BaseCPU *cpu)
//SimObject::takeOverFrom(SimObject *cpu)
{
    //panic("Unimplemented!");
}


SimObject *
SimObject::find(const char *name)
{
    SimObjectList::const_iterator i = simObjectList.begin();
    SimObjectList::const_iterator end = simObjectList.end();

    for (; i != end; ++i) {
        SimObject *obj = *i;
        if (obj->name() == name)
            return obj;
    }

    return NULL;
}
