/*
 * Copyright (c) 2001-2005 The Regents of The University of Michigan
 * Copyright (c) 2010 Advanced Micro Devices, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met: redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer;
 * redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution;
 * neither the name of the copyright holders nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Steve Reinhardt
 *          Nathan Binkert
 */

/* @file
 * User Console Definitions
 */

#ifndef __SIM_OBJECT_HH__
#define __SIM_OBJECT_HH__

#include <iostream>
#include <list>
#include <map>
#include <string>
#include <vector>

#include "params/SimObject.hh"
#include "sim/eventq.hh"
#include "sim/serialize.hh"
#include "base/statistics.hh"
//#include "sim/system.hh"
#include "enums/MemoryMode.hh"

class BaseCPU;
class Event;

void initAll();
void regAllStats();
void SimStartup();

typedef struct {        

   	/*McPAT*/
	//McPAT system stats
	std::vector<std::string> cpuname;
	std::vector<Stats::Counter> total_cycles;
	std::vector<Stats::Counter> idle_cycles;
	std::vector<Stats::Counter> busy_cycles;

	//McPAT core stats
        std::vector<Stats::Counter> total_instructions;
	std::vector<Stats::Counter> int_instructions;
	std::vector<Stats::Counter> fp_instructions;
	std::vector<Stats::Counter> branch_instructions;
	std::vector<Stats::Counter> branch_mispredictions;
	std::vector<Stats::Counter> load_instructions;
	std::vector<Stats::Counter> store_instructions;
	std::vector<Stats::Counter> committed_instructions;
	std::vector<Stats::Counter> committed_int_instructions;
	std::vector<Stats::Counter> committed_fp_instructions;
	std::vector<Stats::Counter> rename_reads;
	std::vector<Stats::Counter> rename_writes;
	std::vector<Stats::Counter> ROB_writes;
	std::vector<Stats::Counter> ROB_reads;
	std::vector<Stats::Counter> fp_rename_reads;
	std::vector<Stats::Counter> fp_rename_writes;

	// --- Instruction window stats
	std::vector<Stats::Counter> inst_window_reads;
	std::vector<Stats::Counter> inst_window_writes;
	std::vector<Stats::Counter> inst_window_wakeup_accesses;
	std::vector<Stats::Counter> fp_inst_window_reads;
	std::vector<Stats::Counter> fp_inst_window_writes;
	std::vector<Stats::Counter> fp_inst_window_wakeup_accesses;

	// --- Register file stats
	std::vector<Stats::Counter> int_regfile_reads;
	std::vector<Stats::Counter> float_regfile_reads;
	std::vector<Stats::Counter> int_regfile_writes;
	std::vector<Stats::Counter> float_regfile_writes;
	std::vector<Stats::Counter> function_calls;
	std::vector<Stats::Counter> context_switches;

	// --- ALU accesses stats
        std::vector<Stats::Counter> ialu_accesses;
        std::vector<Stats::Counter> fpu_accesses;
        std::vector<Stats::Counter> mul_accesses;
        std::vector<Stats::Counter> cdb_alu_accesses;
        std::vector<Stats::Counter> cdb_mul_accesses;
        std::vector<Stats::Counter> cdb_fpu_accesses;

        // McPAT BTB stats
        std::vector<Stats::Counter> BTB_read_accesses;
        std::vector<Stats::Counter> BTB_write_accesses;

/*
	// --- fixed value in McPAT for max power computation 
        float   IFU_duty_cycle;
        float   LSU_duty_cycle;
        float   MemManU_I_duty_cycle;
        float   MemManU_D_duty_cycle;
        float   ALU_duty_cycle;
        float   MUL_duty_cycle;
        float   FPU_duty_cycle;
        float   ALU_cdb_duty_cycle;
        float   MUL_cdb_duty_cycle;
        float   FPU_cdb_duty_cycle;
	float	pipeline_duty_cycle;
	float	l2cache_duty_cycle;
*/

	// McPAT ITLB stats -- for M5 full-system mode
	std::vector<Stats::Counter> itlb_total_accesses;
        std::vector<Stats::Counter> itlb_total_misses;
        std::vector<Stats::Counter> itlb_conflicts;

	// McPAT DTLB stats -- for M5 full-system mode
	std::vector<Stats::Counter> dtlb_total_accesses;
        std::vector<Stats::Counter> dtlb_total_misses;
        std::vector<Stats::Counter> dtlb_conflicts;
		
	// McPAT ICache stats
        std::vector<Stats::Counter> icache_read_accesses;
        std::vector<Stats::Counter> icache_read_misses;
        std::vector<Stats::Counter> icache_conflicts;

        // McPAT DCache stats
        std::vector<Stats::Counter> dcache_read_accesses;
        std::vector<Stats::Counter> dcache_read_misses;
        std::vector<Stats::Counter> dcache_write_accesses;
        std::vector<Stats::Counter> dcache_write_misses;
        std::vector<Stats::Counter> dcache_conflicts;

	// McPAT L2Cache stats
        std::vector<Stats::Counter> l2cache_read_accesses;
        std::vector<Stats::Counter> l2cache_read_misses;
        std::vector<Stats::Counter> l2cache_write_accesses;
        std::vector<Stats::Counter> l2cache_write_misses;
        std::vector<Stats::Counter> l2cache_conflicts;

/*
        // McPAT L3Cache stats
        std::vector<Stats::Counter> l3cache_read_accesses;
        std::vector<Stats::Counter> l3cache_read_misses;
        std::vector<Stats::Counter> l3cache_write_accesses;
        std::vector<Stats::Counter> l3cache_write_misses;
        std::vector<Stats::Counter> l3cache_conflicts;
*/

} usage_counts;


/*
 * Abstract superclass for simulation objects.  Represents things that
 * correspond to physical components and can be specified via the
 * config file (CPUs, caches, etc.).
 */
class SimObject : public EventManager, public Serializable
{
  public:
    enum State {
        Running,
        Draining,
        Drained
    };


  private:
    State state;

  protected:
    void changeState(State new_state) { state = new_state; }

  public:
    State getState() { return state; }

  private:
    typedef std::vector<SimObject *> SimObjectList;

    // list of all instantiated simulation objects
    static SimObjectList simObjectList;
    //static SimObjectList simROIObjectList;
    static SimObjectList simCPUList;
    static SimObjectList simSwitchCPUList;

  protected:
    const SimObjectParams *_params;

  public:
    typedef SimObjectParams Params;
    const Params *params() const { return _params; }
    SimObject(const Params *_params);
    virtual ~SimObject() {}

    int testcounts;

  public:

    virtual const std::string name() const { return params()->name; }

    // The following SimObject initialization methods are called from
    // the instantiate() method in src/python/m5/simulate.py.  See
    // that function for details on how/when these methods are
    // invoked.

    /**
     * init() is called after all C++ SimObjects have been created and
     * all ports are connected.  Initializations that are independent
     * of unserialization but rely on a fully instantiated and
     * connected SimObject graph should be done here.
     */
    virtual void init();
    static void init_All();

    /**
     * loadState() is called on each SimObject when restoring from a
     * checkpoint.  The default implementation simply calls
     * unserialize() if there is a corresponding section in the
     * checkpoint.  However, objects can override loadState() to get
     * other behaviors, e.g., doing other programmed initializations
     * after unserialize(), or complaining if no checkpoint section is
     * found.
     */
    virtual void loadState(Checkpoint *cp);

    /**
     * initState() is called on each SimObject when *not* restoring
     * from a checkpoint.  This provides a hook for state
     * initializations that are only required for a "cold start".
     */
    virtual void initState();

    // register statistics for this object
    virtual void regStats();
    virtual void regFormulas();
    virtual void resetStats();
    virtual void printStats(usage_counts &totalusage);

    // static: call reg_stats on all SimObjects
    static void reg_AllStats();

    // static: call resetStats on all SimObjects
    static void resetAllStats();

    // static: call reg_stats on all SimObjects
    static void printROIStats(usage_counts &totalusage);
    static void printAllStats(usage_counts &totalusage);

    // static: call reg_stats on all SimObjects
    static void SST_FF();

    /**
     * startup() is the final initialization call before simulation.
     * All state is initialized (including unserialized state, if any,
     * such as the curTick() value), so this is the appropriate place to
     * schedule initial event(s) for objects that need them.
     */
    virtual void startup();

    // static: call startup on all SimObjects
    static void Sim_Startup();

    // static: call nameOut() & serialize() on all SimObjects
    static void serializeAll(std::ostream &);

    // Methods to drain objects in order to take checkpoints
    // Or switch from timing -> atomic memory model
    // Drain returns 0 if the simobject can drain immediately or
    // the number of times the drain_event's process function will be called
    // before the object will be done draining. Normally this should be 1
    virtual unsigned int drain(Event *drain_event);
    virtual void resume();
    //virtual void setMemoryMode(State new_mode);
    virtual void setMemoryMode(Enums::MemoryMode mem_mode);
    virtual void switchOut();
    virtual void takeOverFrom(BaseCPU *cpu);
    //virtual void takeOverFrom(SimObject *cpu);

#ifdef DEBUG
  public:
    bool doDebugBreak;
    static void debugObjectBreak(const std::string &objs);
#endif

    /**
     * Find the SimObject with the given name and return a pointer to
     * it.  Primarily used for interactive debugging.  Argument is
     * char* rather than std::string to make it callable from gdb.
     */
    static SimObject *find(const char *name);
};

#endif // __SIM_OBJECT_HH__
